#define EXTERN
#include <mpi.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "hpglobalc.h"
#include "hpsolverc.h"
#include "hpmetrinit.h"

#include "seqfilesm.h"
#include "seqfilest.h"
#include "seqfilesc.h"


int main (int argc,char *argv[])
{
  time_t bt,et;
  long hod,min;
  double sec = clock();

  // start time of the job
  bt = time (NULL);

  //  MPI initialization
  MPI_Init(&argc,&argv);
  MPI_Comm_rank(MPI_COMM_WORLD,&Myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&Nproc);
  MPI_Get_processor_name(proc_namet,&nameLengtht);

  //  initialization of the code
  hpmetr_init (argc,(const char **)(argv));
  
  //  solution of the problem
  solve_metr_problem_parallel ();

  et = time (NULL);
    
  sec = (clock() - sec) / (double)CLOCKS_PER_SEC;
  hod = (long)sec/3600;  sec -= hod*3600;
  min = (long)sec/60;    sec -= min*60;

  if (Myrank==0){
    fprintf (stdout,"\n -------------------------------------------");
    fprintf (stdout,"\n Consumed processor time by HPMETR %2ld:%02ld:%05.2f",hod,min,sec);
    fprintf (stdout,"\n -------------------------------------------\n");

    fprintf (Outc,"\n\n\n Data about computation time \n");
    fprintf (Outc,"\n Total time of computation      %ld", long(et-bt));
    fprintf (Outc,"\n Processor time of computation  %ld:%ld:%.5f", hod, min, sec);

    fprintf (Outc,"\n -------------------------------------------");
    fprintf (Outc,"\n Consumed processor time by HPMETR %ld:%ld:%5f",hod,min,sec);
    fprintf (Outc,"\n -------------------------------------------\n");
  }
    
  fprintf (stdout,"\n");  fprintf (stderr,"\n"); fprintf (Outc,"\n");
  fclose (Out);
  fclose (Outt);
  fclose (Outc);  
  
  MPI_Finalize ();
  
}
